"""Filings downloader and cleaner.

This module downloads filings from the EDGAR database and splits the filings
into multiple files.

"""

import pandas as pd
import multiprocessing as mp
from edgarsearch import filingshandler_worker as fhw
from edgarsearch import tools as t
import re
import uu
import os
import datetime
import itertools
import io

# Local Variables


def multidownload(urls, folder="data/", sub="filings/",
                  edgar_url="https://www.sec.gov/Archives/", **kwargs):
    """Download all filings from a list of URLs.

    Args:
    urls:   List of relative paths to filings in the EDGAR archives
    folder: Path to work working directory (default = data/)
    sub:    Path to subdirectory in working directory (default = filings/)
    edgar_url: URL to archive parent folder on edgar server
               (defaut: "https://www.sec.gov/Archives/")
    Return:
    None

    TODO:
    Make it server friendly with timeouts
    """
    mplist = list(itertools.product(urls, [edgar_url], [folder], [sub]))
    with mp.Pool(mp.cpu_count()) as p:
        result = p.starmap(fhw.singledownload, mplist)
    index = pd.DataFrame(result, columns=["url", "temp_fname", "dt_accessed"])
    return index


def create_filename(row, fname_form="%Y%m%d_%company_", **kwargs):
    """Create filenames for local EDGAR filing documents.

    Args:
    row         Row of Pandas df containing the columns "Date Filed",
                "Company Name" and "url" from the EDGAR index files
    fname_form  String with the filename format.
                Possible variables:
                %org - orignal file name on server
                %company - Company Name
                All date variables based on strftime.
                See: http://strftime.org/
                (default: "%Y%m%d_%company_")
    Return:
    filename as string
    """
    splits = row["url"].split("/")
    org = splits[2] + "_" + splits[3].split(".")[0]

    company = (row["Company Name"]
               .replace(" ", "")
               .replace(",", "")
               .replace(".", "")
               .replace("/", "")
               .replace("\\", "")
               .lower()
               )
    fname = fname_form.replace("%org", org).replace("%company", company)
    tdate = datetime.datetime.strptime(row["Date Filed"], "%Y-%m-%d")
    while fname.find("%") >= 0:
        char = "%" + fname[fname.find("%")+1]
        fname = fname.replace(char, tdate.strftime(char))
    return fname


def splitfiles(docs, folder="data/", sub="filings/", text_only=True, **kwargs):
    """Extract the original documents from the downloaded .txt file from EDGAR.

    Args:
    row         Row from local index Pandas df containing "filename"
    folder      Path of work dir (default = "data/")
    sub         Path of sub dir in work dir (default = "filings/")
    fname_form  String with the filename format.
                Possible variables:
                %org - orignal file name on server
                %company - Company Name
                All date variables based on strftime.
                See: http://strftime.org/
    text_only   If True, only html and txt files are saved. All other types of
                contents such as images and pdfs are discarded
                (default: True)
    Return:     Index of local files
    """
    result = pd.DataFrame(columns=[
                                 "url",
                                 "seq",
                                 "server_fname",
                                 "type",
                                 "desc",
                                 "local_fname",
                                 ])
    for index, row in docs.iterrows():
        tmp_fname = folder + sub + row["temp_fname"]
        fname_part = create_filename(row, **kwargs)
        fname_full = folder + sub + str(fname_part)
        fname_full = t.finduniquefilename(fname_full, exact=False)
        path = fname_full[0:fname_full.rfind("/")]
        if os.path.isdir(path) is False:
            os.makedirs(path)
        f = open(tmp_fname)
        txt = f.read()
        f.close()
        m_sec_header = re.search("<SEC-HEADER>.*</SEC-HEADER>",
                                 txt,
                                 re.DOTALL)
        b_sec_header = m_sec_header.group(0)
        local_fname = fname_full + "header.txt"
        out = open(local_fname, "w")
        out.write(b_sec_header)
        out.close()
        result.loc[result.shape[0]] = ([row["url"],
                                        0,
                                        "HEADER",
                                        "SEC Header",
                                        "Header file",
                                        local_fname])
        starts = ([a.end() for a in list(re.finditer("<DOCUMENT>", txt))])
        ends = ([a.start() for a in list(re.finditer("</DOCUMENT>", txt))])

        for i in range(len(starts)):
            content = txt[starts[i]:ends[i]]
            text = content[content.find("<TEXT>\n")+7:content.find("</TEXT")-1]

            search = ("<TYPE>(?P<TYPE>.+?)[\\n]+?"
                      "<SEQUENCE>(?P<SEQ>.+?)[\\n]+"
                      "<FILENAME>(?P<FNAME>.+?)[\\n]"
                      "(<DESCRIPTION>(?P<DESC>.+?)[\\n])?")
            try:
                info = re.search(search, content, re.MULTILINE).groupdict()
            except:
                print(content)
                break
            if info["TYPE"] != "GRAPHIC":
                local_fname = fname_full + str(info["SEQ"]) + ".html"
                out = open(local_fname, "w")
                out.write(text)
                out.close()
            else:
                if text_only is True:
                    break
                local_fname = fname_full + str(info["SEQ"]) + ".jpg"
                f = io.BytesIO(bytes(text, encoding="ascii"))
                uu.decode(f, local_fname, quiet=True)
                f.close()

            result.loc[result.shape[0]] = ([row["url"],
                                           info["SEQ"],
                                           info["FNAME"],
                                           info["TYPE"],
                                           info["DESC"],
                                           local_fname])
    return result


def replace_img(docs):
    """Replace image path in documents with local filename.

    Args:
    docs:       Pandas df containing all the index of all documents
                belonging to a filing.

    Return:
    None

    """
    for url in docs.url.unique():
        extract_full = docs.loc[docs.url == url]
        extract_graphic = extract_full.loc[extract_full.type == "GRAPHIC"]
        extract_other = extract_full.loc[extract_full.type != "GRAPHIC"]
        if extract_graphic.shape[0] > 0:
            for index, row in extract_other.iterrows():
                f = open(row["local_fname"], "r")
                txt = f.read()
                local = extract_graphic.local_fname.values
                server = extract_graphic.server_fname.values
                for i in range(len(local)):
                    txt = txt.replace(server[i], local[i].split("/")[-1])
                f.close()
                f = open(row["local_fname"], "w")
                f.write(txt)
                f.close


def delfiles(docs, folder="data/", sub="filings/", **kwards):
    """Delete temp raw downloaded files.

    Args:
    row:     Row of Pandas df containing the column "temp_fname"

    Return:
    None
    """
    for index, row in docs.iterrows():
        os.remove(folder + sub + row["temp_fname"])


def batch_process(urls, index, raw=False,
                  text_only=True, chunk_size=100, **kwargs):
    """Execute the defined pipeline in chunks.

    Args:
    urls        List of urls of EDGAR filings to process
    raw         Bool to define wheter the raw filings shall be kept
                or the documents shall be extracted
                (default=False)
    text_only   Bool to defined whether only text files shall be storage or
                non-text files such as images as well
                (default=True)
    chunk_size  Integer for the size of each chunk to process. The bigger the
                chunk, the bigger the temporary file cache.
                (default=100)

    Returns:
    Pandas df containg an index of all raw files
    Pandas df containg an index of all seperated files

    """
    if text_only is False:
        space_req = len(urls) / 1000
        string = ("Saving non-text files may take a lot of space."
                  "The estimated required space is %0.1f GB."
                  " Do you want to continue? [Y]es or [N]o? " % space_req)
        while True:
            check = input(string).lower()
            if ((check == "y") | (check == "yes") |
               (check == "no") | (check == "n")):
                break
        if (check == "no") | (check == "n"):
            return None, None
    if chunk_size > len(urls):
        chunk_size = len(urls)
    c_list = [urls[i:i + chunk_size] for i in range(0, len(urls), chunk_size)]
    final_tmp_f = pd.DataFrame()
    final_docs = pd.DataFrame()
    for urls_sub in c_list:
        tmp_files = (multidownload(urls_sub, **kwargs)
                     .merge(index, left_on="url", right_on="File Name"))
        if raw is False:
            documents = splitfiles(tmp_files, text_only=text_only, **kwargs)
            replace_img(documents)
            delfiles(tmp_files, **kwargs)
            final_docs = pd.concat([final_docs, documents])
        final_tmp_f = pd.concat([final_tmp_f, tmp_files])
    return final_tmp_f.reset_index(), final_docs.reset_index()



"""Filings downloader.

This module downloads filings from the EDGAR database and extracts relevant
information from the filings.

"""
import urllib.request
import uuid
import datetime
import os


def singledownload(url, edgar_url="https://www.sec.gov/Archives/",
                   folder="data/", sub="filings/", **kwargs):
    """Download filings from the EDGAR database.

    Args:
        url: relative path in the EDGAR server
        edgar_url: URL to archive parent folder on edgar server
                   (defaut: "https://www.sec.gov/Archives/")
        folder: Path to work working directory (default = data/)
        sub: Path to subdirectory in working directory (default = filings/)
    Return:
        List of the downloaded url and the relative file path
    """
    full_url = edgar_url + url
    fname = str(uuid.uuid4()) + ".txt"
    txt = urllib.request.urlopen(full_url).read()
    if os.path.isdir(folder+sub) is False:
        os.makedirs(folder+sub)
    try:
        f = open(folder + sub + fname, 'x')
    except:
        f = open(folder + sub + fname, 'w+')
    f.write(txt.decode('utf-8'))
    f.close
    accessed = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    return [url, fname, accessed]



"""EGDAR index downloader.

This module downloads relevant index files from the EGDAR database,
consolidates these. Moreover, filters are applied to only return relevant
results.

Functions in this module:
    download
    consolidates
    filter
"""

import urllib.request
import urllib.error
import os.path
import pandas as pd


def download(sample_start, sample_end, folder, sub):
    """Download index files from the EDGAR database.

    Args:
    sample_start        Start date for filings in sample
                        Datetime objects
    sample_end          End date for filings in sample
                        Datetime objects
    folder              Path to work working directory
    sub                 Path to subdirectory in working directory

    Return:
    None
    """
    url_start = "https://www.sec.gov/Archives/edgar/full-index/"
    url_end = "/form.idx"

    start_excl = []
    if sample_start.month > 3:
        start_excl.append(1)
    if sample_start.month > 6:
        start_excl.append(2)
    if sample_start.month > 9:
        start_excl.append(3)

    end_excl = []
    if sample_end.month < 10:
        end_excl.append(4)
    if sample_end.month < 7:
        end_excl.append(3)
    if sample_end.month < 4:
        end_excl.append(2)

    for year in range(sample_start.year, sample_end.year+1):
        for quarter in range(1, 5):
            if (year == sample_start.year) & (quarter in start_excl):
                continue
            elif (year == sample_end.year) & (quarter in end_excl):
                continue
            else:
                url = url_start + str(year) + "/QTR" + str(quarter) + url_end
                fname = (
                            folder
                            + sub
                            + str(year)
                            + "_"
                            + str(quarter)
                            + ".txt"
                            )
                if os.path.isdir(os.path.split(fname)[0]) is False:
                    os.makedirs(os.path.split(fname)[0])
                if os.path.isfile(fname) is False:
                    try:
                        urllib.request.urlretrieve(url, fname)
                    except urllib.error.HTTPError as e:
                        if e.code == 404:
                            print(("%s not on server!", url))
                        else:
                            print(e)


def consolidate(folder, sub):
    """Consolidate index files and filter for desired form type.

    Args:
    folder              Path to work working directory
    sub                 Path to subdirectory in working directory

    Return:
        Pandas Dataframe of the consolidated index

    """
    index_files = os.listdir(folder + sub)
    result = pd.DataFrame()

    if len(index_files) == 0:
        print("No index files found. Download index files first")
        return None

    for indexf in index_files:
        indexdf = pd.read_fwf(folder + sub + indexf,
                              widths=[12, 62, 12, 12, 52],
                              skiprows=8,
                              comment="---")
        indexdf["date"] = pd.to_datetime(indexdf["Date Filed"])
        result = result.append(indexdf)

    return result


def filter(index, sample_start, sample_end, *,
           filter_formtype=None, filter_CIK=None):
    """Filter index based on form type and/or CIK.

    Args:
    sample_start        Start date for filings in sample
                        Datetime objects
    sample_end          End date for filings in sample
                        Datetime objects
    index               Pandas df containg the full index of the sample period
    filter_formtype     Optional argument to filter based on filings type
                        List with strings e.g. ["8-K, 10-K"]
    filter_CIK          Optional argument to filter based on CIK
                        List with strings e.g. ["12345678", "98765432"]

    Return:
        Pandas Dataframe of the filtered index

    """
    f_period = index.loc[(index.date > sample_start) &
                         (index.date < sample_end)]

    f_type = pd.DataFrame()
    if filter_formtype is not None:
        for formtype in filter_formtype:
            f_type = f_type.append((f_period
                                    .loc[f_period["Form Type"] == formtype]
                                    )
                                   )
    else:
        f_type = index

    f_CIK = pd.DataFrame()
    if filter_CIK is not None:
        for CIK in filter_CIK:
            f_CIK = f_CIK.append(f_type.loc[f_type["CIK"] == CIK])
    else:
        f_CIK = f_type
    return f_CIK



"""Crawl EDGAR database to download index files and filings.

Contains the edgar class.

Sample script:
my_edgar = edgar()
my_edgar.definesearch("20150101", "20151231", filter_formtype=["8-K"])
my_edgar.getindex()
my_edgar.getfilings(text_only=False, sample_size=10,
                    fname_form="%Y/%m/%Y%m_%company")
my_edgar.cur_filings

"""
import datetime
from edgarsearch import indexhandler as ih
from edgarsearch import filingshandler as fh


class edgar:
    """Class for EDGAR searches.

    Contains public functions to download and search index files and also

    """

    def __init__(self,
                 dir_work="edgar/",
                 sub_index="index/",
                 sub_filings="filings/",
                 edgar_url="https://www.sec.gov/Archives/"):
        """Create new EDGAR search object.

        Args:
        dir_work:           Working subdirectory for all saved data
                            (default: "edgar/")
        sub_index:          Subdirectory in dir_work for saved index data
                            (default: "index/")
        sub_filings:        Subdirectory in dir_work for saved filings data
                            (default: "filings")
        edgar_url:          URL to EDGAR server
                            (default: "https://www.sec.gov/Archives/")
        """
        self.dir_work = dir_work
        self.sub_index = sub_index
        self.sub_filings = sub_filings
        self.edgar_url = edgar_url

    def definesearch(self, sample_start, sample_end, *,
                     filter_formtype=None, filter_CIK=None):
        """Set parameters for new search request.

        Args:
        sample_start        Start date for filings in sample
                            String in the format "YYYYMMDD"
        sample_end          End date for filings in sample
                            String in the format "YYYYMMDD"
        filter_formtype     Optional argument to filter based on filings type
                            List with strings e.g. ["8-K, 10-K"]
        filter_CIK          Optional argument to filter based on CIK
                            List with strings e.g. ["12345678", "98765432"]

        """
        try:
            sample_start = datetime.datetime.strptime(sample_start, "%Y%m%d")
            sample_end = datetime.datetime.strptime(sample_end, "%Y%m%d")
        except ValueError:
            print("Dates must be in YYYYMMDD format.")

        if sample_start > sample_end:
            raise ValueError("Sample start must be prior to sample end")
        self.sample_start = sample_start
        self.sample_end = sample_end
        self.filter_formtype = filter_formtype
        self.filter_CIK = filter_CIK

    def getindex(self):
        """Download and store the relevant index.

        Args:
        None

        Return:
        Pandas Dataframe with the corresponding index
        """
        try:
            self.cur_index = ih.filter(self.fullindex,
                                       self.sample_start,
                                       self.sample_end,
                                       filter_formtype=self.filter_formtype,
                                       filter_CIK=self.filter_CIK)
        except:
            ih.download(self.sample_start, self.sample_end,
                        self.dir_work, self.sub_index)
            self.fullindex = ih.consolidate(self.dir_work, self.sub_index)
            self.cur_index = ih.filter(self.fullindex,
                                       self.sample_start,
                                       self.sample_end,
                                       filter_formtype=self.filter_formtype,
                                       filter_CIK=self.filter_CIK)
        return self.cur_index

    def getfilings(self, sample_size=-1, **kwargs):
        """Download the filings matching the search criteria.

        Args:
        sample_size Number of samples to obtain from the selected search.
                    -1 equals the full sample.
                    (default =-1)
        fname_form  String with the filename format.
                    Possible variables:
                    %org - orignal file name on server
                    %company - Company Name
                    All date variables based on strftime.
                    See: http://strftime.org/
                    (default: "%Y%m%d_%company_")
        text_only   If True, only html and txt files are saved. All other types
                    of contents such as images and PDFs are discarded
        raw         Bool to define wheter the raw filings shall be kept
                    or the documents shall be extracted
                    (default=False)
        chunk_size  Integer for the size of each chunk to process. The greater
                    the chunks, the larger the temporary file cache.
                    (default=100)

        Return:
        Pandas DF with the file list
        """
        try:
            if sample_size > 0:
                url_list = (
                            self.cur_index["File Name"]
                            .sample(sample_size)
                            .tolist()
                            )
            else:
                url_list = self.cur_index["File Name"].tolist()
        except:
            try:
                self.cur_index.shape[0] = 0
                print("Current query yields no result.")
            except:
                print("Search not defined or index not yet obtained.")

        tmp_files, *documents = fh.batch_process(url_list,
                                                 self.cur_index,
                                                 folder=self.dir_work,
                                                 sub=self.sub_filings,
                                                 edgar_url=self.edgar_url,
                                                 **kwargs)
        self.tmp_files = tmp_files
        self.cur_filings = documents
        return self.cur_filings



"""Collection of useful tools.

This module contains a collection of useful tools used in a wide variety of
settings.

Currently implemented:
    - Fuzzy-matching of columns from a pandas Dataframe

"""

import os
import string
import glob

def getalpha(x):
    """Convert integer to alphabetic characters.

    Args:
    x           Integer values

    Return:
    String of alphabetic equivalent to integer
    """
    result = ""
    if x // 26 > 0:
        result += getalpha((x // 26)-1)
        result += string.ascii_uppercase[(x % 26)]
    else:
        result += string.ascii_uppercase[(x % 26)]
    return result


def finduniquefilename(fname, mode="alpha", exact=True):
    """Return unique file name if file with original file name already exists.

    Args:
    fname       Filename to be checked
    mode        String setting the mode.
                Possible values:
                "alpha" will add alphabetical suffixes
                "num"  will add numerical suffixes
                default("alpha")
    exact       If true, check for the exact same file will be made
                If false, check whether a file with the same pattern exists

    Return:
    Unique filename
    Augment
    """
    i = 0
    fpre, fsuf = os.path.splitext(fname)
    if exact is True:
        while os.path.isfile(fname):
            if mode == "alpha":
                fname = (
                         fpre
                         + getalpha(i)
                         + fsuf
                         )
            elif mode == "num":
                fname = (
                         fpre
                         + str(i)
                         + fsuf
                         )
            i += 1
    else:
        while len(glob.glob(fname + "*")) > 0:
            if mode == "alpha":
                fname = (
                         fpre
                         + getalpha(i)
                         + fsuf
                         )
            elif mode == "num":
                fname = (
                         fpre
                         + str(i)
                         + fsuf
                         )
            i += 1
    return(fname)



"""Crawl EDGAR database to download index files and filings.

Contains the edgar class.

Sample script:
my_edgar = edgar()
my_edgar.definesearch("20150101", "20151231", filter_formtype=["8-K"])
my_edgar.getindex()
my_edgar.getfilings(text_only=False, sample_size=10,
                    fname_form="%Y/%m/%Y%m_%company")
my_edgar.cur_filings

"""
from setuptools import setup

setup(name="edgarsearch",
      version="0.1",
      description="Crawl EDGAR database to download index files and filings.",
      keywords="EDGAR index filings 8-K 10-K",
      url="https://github.com/markdembo/edgarsearch",
      download_url="https://github.com/markdembo/edgarsearch/archive/0.1.tar.gz",
      author="Mark Dembo",
      author_email="mark.dembo@student.unisg.ch",
      license="MIT",
      packages=["edgarsearch"],
      install_requires=[
          "pandas",
      ],
      )
